const MIN_USERNAME_LENGTH = 3
const MAX_USERNAME_LENGTH = 20
const USERNAME_REGEX = /^[a-z0-9_]+$/

const MIN_PASSWORD_LENGTH = 12
const MAX_PASSWORD_LENGTH = 72
const MIN_PASSWORD_ENTROPY = 60.0

// purposefully produce lower entropy than in the backend to prevent 400s
const ENTROPY_CLASSES: Array<[RegExp, number]> = [
	[/[a-z]/, 24], // 26
	[/[A-Z]/, 24], // 26
	[/\d/, 10],
	[/[!@#$%^&*()\-_+=\[\]{}|;:'",.<>\/?`~\\]/, 32] // 40
]

export const isUsernameValid = (username: string): [boolean, string] => {
	if (username.length < MIN_USERNAME_LENGTH) {
		return [false, `Username cannot be shorter than ${MIN_USERNAME_LENGTH} characters`]
	}

	if (username.length > MAX_USERNAME_LENGTH) {
		return [false, `Username cannot be longer than ${MAX_USERNAME_LENGTH} characters`]
	}

	if (!USERNAME_REGEX.test(username)) {
		return [false, "Username can only contain numbers, letters, and underscores"]
	}

	return [true, ""]
}

export const isPasswordValid = (password: string): [boolean, string] => {
	if (password.length < MIN_PASSWORD_LENGTH) {
		return [false, `Password cannot be shorter than ${MIN_PASSWORD_LENGTH} characters`]
	}

	if (password.length > MAX_PASSWORD_LENGTH) {
		return [false, `Password cannot be longer than ${MAX_PASSWORD_LENGTH} characters`]
	}

	const entropy = calculateEntropy(password)
	console.log(`entropy: ${entropy}`)
	if (entropy < MIN_PASSWORD_ENTROPY) {
		return [
			false,
			"Password is not complex enough (add uppercase, lowercase, numbers, and symbols)"
		]
	}

	return [true, ""]
}

const calculateEntropy = (password: string): number => {
	let poolSize = 0

	for (const [eClass, poolPlus] of ENTROPY_CLASSES) {
		if (eClass.test(password)) {
			poolSize += poolPlus
		}
	}

	// empty password exception
	if (poolSize === 0) {
		return 0
	}

	const uniqueChars = new Set(password.split("")).size

	const basicEntropy = password.length * Math.log2(poolSize)
	const diversityAdjustedEntropy =
		Math.log2(poolSize) + (password.length - 1) * Math.log2(uniqueChars)

	return Math.min(basicEntropy, diversityAdjustedEntropy)
}
